==> sema4ai_actions-0.7.0-py3-none-any.whl: sema4ai/actions/_action.py <==
import inspect
import typing
from contextlib import contextmanager
from typing import Any, Dict, List, Literal, Optional, Tuple, get_type_hints

from robocorp.log import ConsoleMessageKind, console_message
from robocorp.log.protocols import OptExcInfo

from sema4ai.actions._constants import SUPPORTED_TYPES_IN_SCHEMA
from sema4ai.actions._customization._plugin_manager import PluginManager
from sema4ai.actions._protocols import IAction, IContext, Status

_map_python_type_to_user_type = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
}


class Action:
    def __init__(
        self,
        pm: PluginManager,
        module_name: str,
        module_file: str,
        method: typing.Callable,
        options: Optional[Dict] = None,
    ):
        self._pm = pm
        self.module_name = module_name
        self.filename = module_file or "<filename unavailable>"
        self.method = method
        self.message = ""
        self.exc_info: Optional[OptExcInfo] = None
        self._status = Status.NOT_RUN
        self.result = None
        self.options = options or None

    @property
    def name(self):
        return self.method.__code__.co_name

    @property
    def lineno(self):
        return self.method.__code__.co_firstlineno

    def run(self, **kwargs):
        sig = inspect.signature(self.method)
        try:
            sig.bind(**kwargs)
        except Exception as e:
            raise RuntimeError(
                f"It's not possible to call the action: '{self.name}' because the passed arguments don't match the action signature.\nError: {e}"
            )
        return self.method(**kwargs)

    def _build_properties(
        self,
        method_name: str,
        param_name: Optional[str],
        param_type: Any,
        description: str,
        kind: Literal["parameter", "return"],
    ) -> Dict[str, Any]:
        """
        Args:
            method_name: The name of the method for which properties are being built.
            param_name: The parameter name (may be None if it's a return)
            param_type: The type of the property (gotten by introspection).
            description: The description for the property.
            kind: Whether this is a parameter or a return.
        """
        if not param_type:
            param_type_clsname = "string"
        else:
            if param_type not in SUPPORTED_TYPES_IN_SCHEMA:
                if hasattr(param_type, "model_json_schema"):
                    # Support for pydantic
                    from sema4ai.actions._remove_refs import replace_refs

                    # Note: we inline the references and remove the definitions
                    # because this schema can be added as a part of a larger schema
                    # and in doing so the position of the references will reference
                    # an invalid path.
                    ret = replace_refs(param_type.model_json_schema())
                    ret.pop("$defs", None)
                    if description:
                        ret["description"] = description
                    if param_name:
                        ret["title"] = param_name.replace("_", " ").title()
                    return ret
                param_type_clsname = (
                    f"Error. The {kind} type '{param_type.__name__}' in '{method_name}' is not supported. "
                    f"Supported {kind} types: str, int, float, bool or pydantic.BaseModel."
                )
            else:
                param_type_clsname = _map_python_type_to_user_type[param_type]

        properties = {
            "type": param_type_clsname,
            "description": description,
        }

        if param_name:
            properties["title"] = param_name.replace("_", " ").title()
        return properties

    @property
    def managed_params_schema(self) -> Dict[str, Any]:
        from typing import get_args, get_origin

        from sema4ai.actions._commands import _get_managed_param_type, _is_managed_param
        from sema4ai.actions._exceptions import ActionsCollectError

        managed_params_schema: Dict[str, Any] = {}
        sig = inspect.signature(self.method)

        param_name_to_description = self._get_param_name_to_description()
        for param in sig.parameters.values():
            if _is_managed_param(self._pm, param.name, param=param):
                tp = _get_managed_param_type(self._pm, param.name, param=param).__name__

                dct: dict[str, Any] = {"type": tp}
                if tp == "OAuth2Secret":
                    # In this case we need to make some introspection to get additional information
                    # on the type (provider, scopes, ...)
                    annotation_args = get_args(param.annotation)
                    error_message = f"""
Invalid OAuth2Secret annotation found.

The OAuth2Secret must be parametrized with 2 arguments, 
the first being a Literal with the provider name 
(i.e.: `Literal["google"]`) 
and the second a list with one Literal with the (multiple) scopes that are required
(i.e.: `list[Literal["scope1", "scope2", ...]])

Full Example for a parameter named `google_secret` requiring google sheets 
and google drive scopes:

google_secret: OAuth2Secret[
    Literal["google"],
    list[
        Literal[
            "https://www.googleapis.com/auth/drive.readonly",
            "https://www.googleapis.com/auth/spreadsheets.readonly",
        ]
    ],
]

File with @action: {self.filename}
@action name: {self.name}
"""
                    if not annotation_args or len(annotation_args) != 2:
                        raise ActionsCollectError(error_message)

                    provider = annotation_args[0]
                    scopes = annotation_args[1]
                    if get_origin(provider) != Literal:
                        raise ActionsCollectError(
                            f"First parameter is not a Literal.\n{error_message}"
                        )
                    if get_origin(scopes) not in (list, List):
                        raise ActionsCollectError(
                            f"Second parameter is not a list.\n{error_message}"
                        )
                    provider_args = get_args(provider)
                    if not provider_args or len(provider_args) != 1:
                        raise ActionsCollectError(
                            f"First parameter does not have a single provider argument.\n{error_message}"
                        )
                    provider_str = provider_args[0]
                    if not isinstance(provider_str, str):
                        raise ActionsCollectError(
                            f"First parameter Literal does not have a string.\n{error_message}"
                        )

                    dct["provider"] = provider_str

                    scope_args = get_args(scopes)
                    if not scope_args or len(scope_args) != 1:
                        raise ActionsCollectError(
                            f"Second parameter is not a list with a single Literal.\n{error_message}"
                        )

                    scope_args_literal = scope_args[0]
                    if get_origin(scope_args_literal) != Literal:
                        raise ActionsCollectError(
                            f"Second parameter is not a list with a single Literal.\n{error_message}"
                        )

                    scope_strs = get_args(scope_args_literal)
                    for entry in scope_strs:
                        if not isinstance(entry, str):
                            raise ActionsCollectError(
                                f"Second parameter is not a list with a literal with strings.\n{error_message}"
                            )

                    dct["scopes"] = scope_strs

                desc = param_name_to_description.get(param.name)
                if desc:
                    dct["description"] = desc
                managed_params_schema[param.name] = dct

        return managed_params_schema

    def _get_param_name_to_description(self) -> dict[str, str]:
        import docstring_parser

        param_name_to_description: Dict[str, str] = {}
        doc = getattr(self.method, "__doc__", "")
        if doc:
            contents = docstring_parser.parse(doc)
            for docparam in contents.params:
                if docparam.description:
                    param_name_to_description[docparam.arg_name] = docparam.description

        return param_name_to_description

    @property
    def input_schema(self) -> Dict[str, Any]:
        from sema4ai.actions._commands import _is_managed_param

        sig = inspect.signature(self.method)
        method_name = self.method.__code__.co_name
        type_hints = get_type_hints(self.method)

        param_name_to_description: dict[
            str, str
        ] = self._get_param_name_to_description()

        properties: Dict[str, Any] = {}
        required: List[str] = []

        schema = {
            "properties": properties,
            "type": "object",
        }

        for param in sig.parameters.values():
            if _is_managed_param(self._pm, param.name, param=param):
                continue
            param_type = type_hints.get(param.name)
            description = param_name_to_description.get(param.name, "")
            param_properties = self._build_properties(
                method_name, param.name, param_type, description, "parameter"
            )
            properties[param.name] = param_properties

            if param.default is inspect.Parameter.empty:
                required.append(param.name)
            else:
                param_properties["default"] = param.default

        if required:
            schema["required"] = required

        return schema

    @property
    def output_schema(self) -> Dict[str, Any]:
        import docstring_parser

        method_name = self.method.__code__.co_name
        type_hints = get_type_hints(self.method)

        doc = getattr(self.method, "__doc__", "")
        description = ""
        if doc:
            contents = docstring_parser.parse(doc)
            returns = contents.returns
            if returns and returns.description:
                description = returns.description

        schema = self._build_properties(
            method_name, None, type_hints.get("return"), description, "return"
        )
        return schema

    @property
    def status(self) -> Status:
        return self._status

    @status.setter
    def status(self, value: Status):
        self._status = Status(value)

    @property
    def failed(self):
        return self._status == Status.FAIL

    def __typecheckself__(self) -> None:
        from sema4ai.actions._protocols import check_implements

        _: IAction = check_implements(self)

    def __str__(self):
        return f"Action({self.name}, status: {self.status})"

    __repr__ = __str__


class _ActionContext:
    _current_action: Optional[IAction] = None


def set_current_action(action: Optional[IAction]):
    _ActionContext._current_action = action


def get_current_action() -> Optional[IAction]:
    return _ActionContext._current_action


class Context:
    # Some regular message (generated by the framework).
    KIND_REGULAR = ConsoleMessageKind.REGULAR

    # Some message which deserves a bit more attention (generated by the framework).
    KIND_IMPORTANT = ConsoleMessageKind.IMPORTANT

    # The action name is being written (generated by the framework).
    KIND_TASK_NAME = ConsoleMessageKind.TASK_NAME

    # Some error message (generated by the framework).
    KIND_ERROR = ConsoleMessageKind.ERROR

    # Some traceback message (generated by the framework).
    KIND_TRACEBACK = ConsoleMessageKind.TRACEBACK

    # Some user message which was being sent to the stdout.
    KIND_STDOUT = ConsoleMessageKind.STDOUT

    # Some user message which was being sent to the stderr.
    KIND_STDERR = ConsoleMessageKind.STDERR

    def __init__(self, print_result: bool = False):
        self._msg_len_target = 80
        self._print_result = print_result

    def _show_header(self, parts: List[Tuple[str, str]]) -> int:
        """
        Returns:
            The final number of chars used in the full header.
        """
        if not parts:
            self.show("=" * self._msg_len_target)
            return self._msg_len_target

        total_len = sum(len(msg) for (msg, _) in parts)

        # + 2 for the spacing before after the '='.
        diff = self._msg_len_target - (total_len + 2)

        start_sep_chars = 3
        end_sep_chars = 3
        if diff > 0:
            start_sep_chars = int(diff / 2)
            end_sep_chars = diff - start_sep_chars

        final_len = 0
        msg_start = f"{'=' * start_sep_chars} "
        final_len += len(msg_start)
        self.show(msg_start, end="")

        for msg, kind in parts:
            final_len += len(msg)
            self.show(msg, end="", kind=kind)

        msg_end = f" {'=' * end_sep_chars}"
        final_len += len(msg_end)
        self.show(msg_end)

        return final_len

    def show(
        self, msg: str, end: str = "\n", kind=KIND_REGULAR, flush: Optional[bool] = None
    ):
        """
        Shows a message to the user.

        Args:
            msg: The message to be shown.
            end: The end char to be added to the message.
            kind: The kind of the message.
            flush: Whether we should flush after sending the message (if None
                   it's flushed if the end char ends with '\n').
        """
        if end:
            msg = f"{msg}{end}"
        console_message(msg, kind, flush=flush)

    def show_error(self, msg: str, flush: Optional[bool] = None):
        self.show(msg, kind=self.KIND_ERROR, flush=flush)

    def _before_action_run(self, action: IAction):
        self._msg_len_target = 80
        self._msg_len_target = self._show_header(
            [("Running: ", self.KIND_REGULAR), (action.name, self.KIND_TASK_NAME)]
        )

    def _after_action_run(self, action: IAction):
        import json
        import traceback

        msg = ""
        if action.message:
            msg = f"\n{action.message}"

        status_kind = self.KIND_REGULAR
        if action.status == Status.FAIL:
            status_kind = self.KIND_ERROR

        show = self.show

        result = action.result

        show(f"{action.name}", end="", kind=self.KIND_TASK_NAME)
        show(" status: ", end="")
        show(f"{action.status.value}", kind=status_kind)

        if msg:
            show(f"{msg}", kind=status_kind)

            if action.exc_info and action.exc_info[0]:
                self._show_header(
                    [
                        ("Full Traceback (running ", self.KIND_REGULAR),
                        (action.name, self.KIND_TASK_NAME),
                        (")", self.KIND_REGULAR),
                    ]
                )

                self.show(
                    "".join(traceback.format_exception(*action.exc_info)),
                    kind=self.KIND_TRACEBACK,
                )

        if self._print_result and action.status == Status.PASS:
            show("result:", kind=self.KIND_IMPORTANT)
            try:
                if hasattr(result, "model_dump_json"):
                    # Support for pydantic
                    result_as_json_str = result.model_dump_json(indent=4)
                else:
                    result_as_json_str = json.dumps(result, indent=4)
            except Exception:
                raise RuntimeError(
                    "The resulting value: {result} cannot be converted to JSON."
                )

            show(
                result_as_json_str,
                flush=True,
            )

        self._show_header([])

    @contextmanager
    def register_lifecycle_prints(self):
        from sema4ai.actions._hooks import after_action_run, before_action_run

        with before_action_run.register(
            self._before_action_run
        ), after_action_run.register(self._after_action_run):
            yield

    def __typecheckself__(self) -> None:
        from sema4ai.actions._protocols import check_implements

        _: IContext = check_implements(self)

==> sema4ai_actions-0.8.0-py3-none-any.whl: sema4ai/actions/_action.py <==
import inspect
import typing
from contextlib import contextmanager
from typing import Any, Dict, List, Literal, Optional, Tuple, get_type_hints

from robocorp.log import ConsoleMessageKind, console_message
from robocorp.log.protocols import OptExcInfo

from sema4ai.actions._constants import SUPPORTED_TYPES_IN_SCHEMA
from sema4ai.actions._customization._plugin_manager import PluginManager
from sema4ai.actions._protocols import IAction, IContext, Status

_map_python_type_to_user_type = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
}


class Action:
    def __init__(
        self,
        pm: PluginManager,
        module_name: str,
        module_file: str,
        method: typing.Callable,
        options: Optional[Dict] = None,
    ):
        self._pm = pm
        self.module_name = module_name
        self.filename = module_file or "<filename unavailable>"
        self.method = method
        self.message = ""
        self.exc_info: Optional[OptExcInfo] = None
        self._status = Status.NOT_RUN
        self.result = None
        self.options = options or None

    @property
    def name(self):
        return self.method.__code__.co_name

    @property
    def lineno(self):
        return self.method.__code__.co_firstlineno

    def run(self, **kwargs):
        sig = inspect.signature(self.method)
        try:
            sig.bind(**kwargs)
        except Exception as e:
            raise RuntimeError(
                f"It's not possible to call the action: '{self.name}' because the passed arguments don't match the action signature.\nError: {e}"
            )
        return self.method(**kwargs)

    def _build_properties(
        self,
        method_name: str,
        param_name: Optional[str],
        param_type: Any,
        description: str,
        kind: Literal["parameter", "return"],
    ) -> Dict[str, Any]:
        """
        Args:
            method_name: The name of the method for which properties are being built.
            param_name: The parameter name (may be None if it's a return)
            param_type: The type of the property (gotten by introspection).
            description: The description for the property.
            kind: Whether this is a parameter or a return.
        """
        if not param_type:
            param_type_clsname = "string"
        else:
            if param_type not in SUPPORTED_TYPES_IN_SCHEMA:
                if hasattr(param_type, "model_json_schema"):
                    # Support for pydantic
                    from sema4ai.actions._remove_refs import replace_refs

                    # Note: we inline the references and remove the definitions
                    # because this schema can be added as a part of a larger schema
                    # and in doing so the position of the references will reference
                    # an invalid path.
                    ret = replace_refs(param_type.model_json_schema())
                    ret.pop("$defs", None)
                    if description:
                        ret["description"] = description
                    if param_name:
                        ret["title"] = param_name.replace("_", " ").title()
                    return ret
                param_type_clsname = (
                    f"Error. The {kind} type '{param_type.__name__}' in '{method_name}' is not supported. "
                    f"Supported {kind} types: str, int, float, bool or pydantic.BaseModel."
                )
            else:
                param_type_clsname = _map_python_type_to_user_type[param_type]

        properties = {
            "type": param_type_clsname,
            "description": description,
        }

        if param_name:
            properties["title"] = param_name.replace("_", " ").title()
        return properties

    @property
    def managed_params_schema(self) -> Dict[str, Any]:
        from typing import get_args, get_origin

        from sema4ai.actions._commands import _get_managed_param_type, _is_managed_param
        from sema4ai.actions._exceptions import ActionsCollectError

        managed_params_schema: Dict[str, Any] = {}
        sig = inspect.signature(self.method)

        param_name_to_description = self._get_param_name_to_description()
        for param in sig.parameters.values():
            if _is_managed_param(self._pm, param.name, param=param):
                tp = _get_managed_param_type(self._pm, param.name, param=param).__name__

                dct: dict[str, Any] = {"type": tp}
                if tp == "OAuth2Secret":
                    # In this case we need to make some introspection to get additional information
                    # on the type (provider, scopes, ...)
                    annotation_args = get_args(param.annotation)
                    error_message = f"""
Invalid OAuth2Secret annotation found.

The OAuth2Secret must be parametrized with 2 arguments, 
the first being a Literal with the provider name 
(i.e.: `Literal["google"]`) 
and the second a list with one Literal with the (multiple) scopes that are required
(i.e.: `list[Literal["scope1", "scope2", ...]])

Full Example for a parameter named `google_secret` requiring google sheets 
and google drive scopes:

google_secret: OAuth2Secret[
    Literal["google"],
    list[
        Literal[
            "https://www.googleapis.com/auth/drive.readonly",
            "https://www.googleapis.com/auth/spreadsheets.readonly",
        ]
    ],
]

File with @action: {self.filename}
@action name: {self.name}
"""
                    if not annotation_args or len(annotation_args) != 2:
                        raise ActionsCollectError(error_message)

                    provider = annotation_args[0]
                    scopes = annotation_args[1]
                    if get_origin(provider) != Literal:
                        raise ActionsCollectError(
                            f"First parameter is not a Literal.\n{error_message}"
                        )
                    if get_origin(scopes) not in (list, List):
                        raise ActionsCollectError(
                            f"Second parameter is not a list.\n{error_message}"
                        )
                    provider_args = get_args(provider)
                    if not provider_args or len(provider_args) != 1:
                        raise ActionsCollectError(
                            f"First parameter does not have a single provider argument.\n{error_message}"
                        )
                    provider_str = provider_args[0]
                    if not isinstance(provider_str, str):
                        raise ActionsCollectError(
                            f"First parameter Literal does not have a string.\n{error_message}"
                        )

                    dct["provider"] = provider_str

                    scope_args = get_args(scopes)
                    if not scope_args or len(scope_args) != 1:
                        raise ActionsCollectError(
                            f"Second parameter is not a list with a single Literal.\n{error_message}"
                        )

                    scope_args_literal = scope_args[0]
                    if get_origin(scope_args_literal) != Literal:
                        raise ActionsCollectError(
                            f"Second parameter is not a list with a single Literal.\n{error_message}"
                        )

                    scope_strs = get_args(scope_args_literal)
                    for entry in scope_strs:
                        if not isinstance(entry, str):
                            raise ActionsCollectError(
                                f"Second parameter is not a list with a literal with strings.\n{error_message}"
                            )

                    dct["scopes"] = scope_strs

                desc = param_name_to_description.get(param.name)
                if desc:
                    dct["description"] = desc
                managed_params_schema[param.name] = dct

        return managed_params_schema

    def _get_param_name_to_description(self) -> dict[str, str]:
        import docstring_parser

        param_name_to_description: Dict[str, str] = {}
        doc = getattr(self.method, "__doc__", "")
        if doc:
            contents = docstring_parser.parse(doc)
            for docparam in contents.params:
                if docparam.description:
                    param_name_to_description[docparam.arg_name] = docparam.description

        return param_name_to_description

    @property
    def input_schema(self) -> Dict[str, Any]:
        from sema4ai.actions._commands import _is_managed_param

        sig = inspect.signature(self.method)
        method_name = self.method.__code__.co_name
        type_hints = get_type_hints(self.method)

        param_name_to_description: dict[
            str, str
        ] = self._get_param_name_to_description()

        properties: Dict[str, Any] = {}
        required: List[str] = []

        schema = {
            "properties": properties,
            "type": "object",
        }

        for param in sig.parameters.values():
            if _is_managed_param(self._pm, param.name, param=param):
                continue
            param_type = type_hints.get(param.name)
            description = param_name_to_description.get(param.name, "")
            param_properties = self._build_properties(
                method_name, param.name, param_type, description, "parameter"
            )
            properties[param.name] = param_properties

            if param.default is inspect.Parameter.empty:
                required.append(param.name)
            else:
                param_properties["default"] = param.default

        if required:
            schema["required"] = required

        return schema

    @property
    def output_schema(self) -> Dict[str, Any]:
        import docstring_parser

        method_name = self.method.__code__.co_name
        type_hints = get_type_hints(self.method)

        doc = getattr(self.method, "__doc__", "")
        description = ""
        if doc:
            contents = docstring_parser.parse(doc)
            returns = contents.returns
            if returns and returns.description:
                description = returns.description

        schema = self._build_properties(
            method_name, None, type_hints.get("return"), description, "return"
        )
        return schema

    @property
    def status(self) -> Status:
        return self._status

    @status.setter
    def status(self, value: Status):
        self._status = Status(value)

    @property
    def failed(self):
        return self._status == Status.FAIL

    def __typecheckself__(self) -> None:
        from sema4ai.actions._protocols import check_implements

        _: IAction = check_implements(self)

    def __str__(self):
        return f"Action({self.name}, status: {self.status})"

    __repr__ = __str__


class _ActionContext:
    _current_action: Optional[IAction] = None


def set_current_action(action: Optional[IAction]):
    _ActionContext._current_action = action


def get_current_action() -> Optional[IAction]:
    return _ActionContext._current_action


class Context:
    # Some regular message (generated by the framework).
    KIND_REGULAR = ConsoleMessageKind.REGULAR

    # Some message which deserves a bit more attention (generated by the framework).
    KIND_IMPORTANT = ConsoleMessageKind.IMPORTANT

    # The action name is being written (generated by the framework).
    KIND_TASK_NAME = ConsoleMessageKind.TASK_NAME

    # Some error message (generated by the framework).
    KIND_ERROR = ConsoleMessageKind.ERROR

    # Some traceback message (generated by the framework).
    KIND_TRACEBACK = ConsoleMessageKind.TRACEBACK

    # Some user message which was being sent to the stdout.
    KIND_STDOUT = ConsoleMessageKind.STDOUT

    # Some user message which was being sent to the stderr.
    KIND_STDERR = ConsoleMessageKind.STDERR

    def __init__(self, print_result: bool = False):
        self._msg_len_target = 80
        self._print_result = print_result

    def _show_header(self, parts: List[Tuple[str, str]]) -> int:
        """
        Returns:
            The final number of chars used in the full header.
        """
        if not parts:
            self.show("=" * self._msg_len_target)
            return self._msg_len_target

        total_len = sum(len(msg) for (msg, _) in parts)

        # + 2 for the spacing before after the '='.
        diff = self._msg_len_target - (total_len + 2)

        start_sep_chars = 3
        end_sep_chars = 3
        if diff > 0:
            start_sep_chars = int(diff / 2)
            end_sep_chars = diff - start_sep_chars

        final_len = 0
        msg_start = f"{'=' * start_sep_chars} "
        final_len += len(msg_start)
        self.show(msg_start, end="")

        for msg, kind in parts:
            final_len += len(msg)
            self.show(msg, end="", kind=kind)

        msg_end = f" {'=' * end_sep_chars}"
        final_len += len(msg_end)
        self.show(msg_end)

        return final_len

    def show(
        self, msg: str, end: str = "\n", kind=KIND_REGULAR, flush: Optional[bool] = None
    ):
        """
        Shows a message to the user.

        Args:
            msg: The message to be shown.
            end: The end char to be added to the message.
            kind: The kind of the message.
            flush: Whether we should flush after sending the message (if None
                   it's flushed if the end char ends with '\n').
        """
        if end:
            msg = f"{msg}{end}"
        console_message(msg, kind, flush=flush)

    def show_error(self, msg: str, flush: Optional[bool] = None):
        self.show(msg, kind=self.KIND_ERROR, flush=flush)

    def _before_action_run(self, action: IAction):
        self._msg_len_target = 80
        self._msg_len_target = self._show_header(
            [("Running: ", self.KIND_REGULAR), (action.name, self.KIND_TASK_NAME)]
        )

    def _after_action_run(self, action: IAction):
        import json
        import traceback

        msg = ""
        if action.message:
            msg = f"\n{action.message}"

        status_kind = self.KIND_REGULAR
        if action.status == Status.FAIL:
            status_kind = self.KIND_ERROR

        show = self.show

        result = action.result

        show(f"{action.name}", end="", kind=self.KIND_TASK_NAME)
        show(" status: ", end="")
        show(f"{action.status.value}", kind=status_kind)

        if msg:
            show(f"{msg}", kind=status_kind)

            if action.exc_info and action.exc_info[0]:
                self._show_header(
                    [
                        ("Full Traceback (running ", self.KIND_REGULAR),
                        (action.name, self.KIND_TASK_NAME),
                        (")", self.KIND_REGULAR),
                    ]
                )

                self.show(
                    "".join(traceback.format_exception(*action.exc_info)),
                    kind=self.KIND_TRACEBACK,
                )

        if self._print_result and action.status == Status.PASS:
            show("result:", kind=self.KIND_IMPORTANT)
            try:
                dump = result
                if hasattr(result, "model_dump"):
                    # Support for pydantic
                    dump = result.model_dump()

                result_as_json_str = json.dumps(dump, indent=4)
            except Exception:
                raise RuntimeError(
                    "The resulting value: {result} cannot be converted to JSON."
                )

            show(
                result_as_json_str,
                flush=True,
            )

        self._show_header([])

    @contextmanager
    def register_lifecycle_prints(self):
        from sema4ai.actions._hooks import after_action_run, before_action_run

        with before_action_run.register(
            self._before_action_run
        ), after_action_run.register(self._after_action_run):
            yield

    def __typecheckself__(self) -> None:
        from sema4ai.actions._protocols import check_implements

        _: IContext = check_implements(self)
